import { BackHandler } from 'react-native';
import { invariant, getStackItemOptions, defaultOptions } from '../utils';

const createModalState = () => {
  let state = {
    currentModal: null,
    stack: {
      names: [],
      content: [],
      defaultOptions,
      openedItems: new Set(),
      pendingClosingActions: new Set()
    }
  };
  const stateListeners = new Set();

  const setState = updater => {
    state = updater(getState());
    stateListeners.forEach(stateListener => stateListener());
    return state;
  };

  const init = setState;

  const getState = () => state;

  const addStateSubscriber = stateSubscriber => {
    function stateListener() {
      try {
        const currentState = getState();

        if (!stateSubscriber.equalityFn(stateSubscriber.state, currentState)) {
          stateSubscriber.stateListener(stateSubscriber.state = currentState);
        }
      } catch (error) {
        stateSubscriber.error = true;
        stateSubscriber.stateListener(null, error);
      }
    }

    stateListeners.add(stateListener);
    return {
      unsubscribe: () => stateListeners.delete(stateListener)
    };
  };

  const createStateSubscriber = (stateListener, equalityFn) => ({
    equalityFn,
    error: false,
    stateListener,
    state: getState(),
    unsubscribe: () => true
  });

  const subscribe = function (stateListener) {
    let equalityFn = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : Object.is;
    return addStateSubscriber(createStateSubscriber(stateListener, equalityFn));
  };

  const openModal = _ref => {
    let {
      modalName,
      params,
      isCalledOutsideOfContext,
      callback
    } = _ref;
    const {
      currentModal,
      stack: {
        content,
        names
      }
    } = getState();
    invariant(modalName, "You didn't pass any modal name");
    invariant(names.some(name => name === modalName), `'${modalName}' is not a valid modal name. Did you mean any of these: ${names.map(validName => `\n• ${validName}`)}`);
    const stackItem = content.find(item => item.name === modalName);
    const hash = `${modalName}_${Math.random().toString(36).substring(2, 11)}`;

    if (!currentModal && isCalledOutsideOfContext) {
      BackHandler.addEventListener('hardwareBackPress', handleBackPress);
    }

    if (stackItem) {
      setState(currentState => ({
        currentModal: modalName,
        stack: { ...currentState.stack,
          openedItems: currentState.stack.openedItems.add(Object.assign({}, stackItem, { ...stackItem,
            hash,
            callback,
            ...(params && {
              params
            })
          }))
        }
      }));
    }
  };

  const getParam = (hash, paramName, defaultValue) => {
    var _stackItem$params$par, _stackItem, _stackItem$params;

    const {
      stack: {
        openedItems
      }
    } = getState();
    let stackItem;
    openedItems.forEach(item => {
      if (item.hash === hash) stackItem = item;
    });
    return (_stackItem$params$par = (_stackItem = stackItem) === null || _stackItem === void 0 ? void 0 : (_stackItem$params = _stackItem.params) === null || _stackItem$params === void 0 ? void 0 : _stackItem$params[paramName]) !== null && _stackItem$params$par !== void 0 ? _stackItem$params$par : defaultValue;
  };

  const closeModal = closingElement => {
    const {
      stack: {
        openedItems: oldOpenedItems,
        names
      }
    } = getState();
    const newOpenedItems = new Set(oldOpenedItems);

    if (typeof closingElement === 'string') {
      invariant(names.some(name => name === closingElement), `'${closingElement}' is not a valid modal name. Did you mean any of these: ${names.map(validName => `\n• ${String(validName)}`)}`);
      let wasItemRemoved = false;
      const reversedOpenedItemsArray = Array.from(newOpenedItems).reverse();
      reversedOpenedItemsArray.forEach(openedItem => {
        if (openedItem.name === closingElement && !wasItemRemoved) {
          newOpenedItems.delete(openedItem);
          wasItemRemoved = true;
        }
      });

      if (!wasItemRemoved) {
        console.warn(`There was no opened ${closingElement} modal.`);
      } // @ts-ignore

    } else if (closingElement && newOpenedItems.has(closingElement)) {
      // @ts-ignore
      newOpenedItems.delete(closingElement);
    } else {
      const staleStackItem = Array.from(newOpenedItems).pop();
      if (staleStackItem) newOpenedItems.delete(staleStackItem);
    }

    const newOpenedItemsArray = Array.from(newOpenedItems);
    setState(currentState => {
      var _newOpenedItemsArray;

      return {
        stack: { ...currentState.stack,
          openedItems: newOpenedItems
        },
        currentModal: newOpenedItemsArray === null || newOpenedItemsArray === void 0 ? void 0 : (_newOpenedItemsArray = newOpenedItemsArray[(newOpenedItemsArray === null || newOpenedItemsArray === void 0 ? void 0 : newOpenedItemsArray.length) - 1]) === null || _newOpenedItemsArray === void 0 ? void 0 : _newOpenedItemsArray.name
      };
    });
  };

  const closeModals = modalName => {
    const {
      stack: {
        openedItems: oldOpenedItems,
        names
      }
    } = getState();
    invariant(modalName, "You didn't pass any modal name to closeModals()");
    invariant(names.some(name => name === modalName), `'${modalName}' is not a valid modal name. Did you mean any of these: ${names.map(validName => `\n• ${String(validName)}`)}`);
    const newOpenedItems = new Set(oldOpenedItems);
    newOpenedItems.forEach(item => {
      if (item.name === modalName) newOpenedItems.delete(item);
    });

    if (newOpenedItems.size !== oldOpenedItems.size) {
      const openedItemsArray = Array.from(newOpenedItems);
      setState(currentState => {
        var _openedItemsArray;

        return {
          stack: { ...currentState.stack,
            openedItems: newOpenedItems
          },
          currentModal: openedItemsArray === null || openedItemsArray === void 0 ? void 0 : (_openedItemsArray = openedItemsArray[(openedItemsArray === null || openedItemsArray === void 0 ? void 0 : openedItemsArray.length) - 1]) === null || _openedItemsArray === void 0 ? void 0 : _openedItemsArray.name
        };
      });
      return true;
    }

    return false;
  };

  const closeAllModals = () => {
    setState(currentState => ({
      currentModal: null,
      stack: { ...currentState.stack,
        openedItems: new Set()
      }
    }));
  };

  const handleBackPress = () => {
    const {
      currentModal,
      stack
    } = getState();
    const currentModalStackItem = Array.from(stack.openedItems).pop();
    const {
      backBehavior
    } = getStackItemOptions(currentModalStackItem, stack);

    if (currentModal) {
      if (backBehavior === 'none') return true;else if (backBehavior === 'clear') {
        queueClosingAction({
          action: 'closeAllModals'
        });
        return true;
      } else if (backBehavior === 'pop') {
        queueClosingAction({
          action: 'closeModal',
          modalName: currentModal
        });
        return true;
      }
    }

    return false;
  };

  const queueClosingAction = _ref2 => {
    let {
      action,
      callback,
      modalName
    } = _ref2;
    const {
      stack: {
        names,
        openedItems
      }
    } = getState();

    if (action !== 'closeAllModals' && modalName) {
      invariant(names.some(name => name === modalName), `'${modalName}' is not a valid modal name. Did you mean any of these: ${names.map(validName => `\n• ${validName}`)}`);
    }

    const noOpenedItems = !(openedItems !== null && openedItems !== void 0 && openedItems.size);
    if (noOpenedItems) return null;
    const hash = `${modalName ? `${modalName}_${action}` : action}_${Math.random().toString(36).substring(2, 11)}`;
    const {
      pendingClosingActions
    } = setState(currentState => {
      var _slice$;

      const newPendingClosingActions = new Set(currentState.stack.pendingClosingActions);
      newPendingClosingActions.add({
        hash,
        action,
        callback,
        modalName,
        currentModalHash: (_slice$ = [...currentState.stack.openedItems].slice(-1)[0]) === null || _slice$ === void 0 ? void 0 : _slice$.hash
      });
      return { ...currentState,
        stack: { ...currentState.stack,
          pendingClosingActions: newPendingClosingActions
        }
      };
    }).stack;
    return [...pendingClosingActions].slice(-1)[0];
  };

  const removeClosingAction = action => {
    const {
      stack: {
        pendingClosingActions: oldPendingClosingActions
      }
    } = getState();
    const newPendingClosingActions = new Set(oldPendingClosingActions);

    if (newPendingClosingActions.has(action)) {
      newPendingClosingActions.delete(action);
    } else return false;

    setState(currentState => ({ ...currentState,
      stack: { ...currentState.stack,
        pendingClosingActions: newPendingClosingActions
      }
    }));
    return true;
  };

  return {
    init,
    setState,
    getState,
    getParam,
    openModal,
    subscribe,
    closeModal,
    closeModals,
    closeAllModals,
    handleBackPress,
    queueClosingAction,
    removeClosingAction
  };
};

const ModalState = createModalState();
/**
 * Function that exposes Modalfy's API outside of React's context.
 *
 * Note: Do not use if you're inside a React component.
 * Please consider `useModal()` or `withModal()` instead.
 *
 * @returns Object containing all the functions and variables of the usual `modal` prop.
 *
 * @see https://colorfy-software.gitbook.io/react-native-modalfy/api/modalfy
 */

export const modalfy = () => {
  var _ModalState$getState$, _ModalState$getState;

  return {
    /**
     * This function closes every open modal.
     *
     * @example modalfy().closeAllModals(() => console.log('All modals closed'))
     *
     * @see https://colorfy-software.gitbook.io/react-native-modalfy/api/types/modalprop#closeallmodals
     */
    closeAllModals: callback => {
      ModalState.queueClosingAction({
        action: 'closeAllModals',
        callback
      });
    },

    /**
     * This function closes the currently displayed modal by default.
     *
     * You can also provide a `modalName` if you want to close a different modal
     * than the latest opened. This will only close the latest instance of that modal,
     * see `closeModals()` if you want to close all instances.
     *
     * @example modalfy().closeModal('ExampleModal', () => console.log('Current modal closed'))
     *
     * @see https://colorfy-software.gitbook.io/react-native-modalfy/api/types/modalprop#closemodal
     */
    closeModal: (modalName, callback) => {
      ModalState.queueClosingAction({
        action: 'closeModal',
        modalName,
        callback
      });
    },

    /**
     * This function closes all the instances of a given modal.
     *
     * You can use it whenever you have the same modal opened
     * several times, to close all of them at once.
     *
     * @example modalfy().closeModals('ExampleModal', () => console.log('All ExampleModal modals closed'))
     *
     * @returns { boolean } Whether or not Modalfy found any open modal
     * corresponding to `modalName` (and then closed them).
     *
     * @see https://colorfy-software.gitbook.io/react-native-modalfy/api/types/modalprop#closemodals
     */
    closeModals: (modalName, callback) => {
      ModalState.queueClosingAction({
        action: 'closeModals',
        modalName,
        callback
      });
    },

    /**
     * This value returns the current open modal (`null` if none).
     *
     * @example modalfy().currentModal
     *
     * @see https://colorfy-software.gitbook.io/react-native-modalfy/api/types/modalprop#currentmodal
     */
    currentModal: (_ModalState$getState$ = (_ModalState$getState = ModalState.getState()) === null || _ModalState$getState === void 0 ? void 0 : _ModalState$getState.currentModal) !== null && _ModalState$getState$ !== void 0 ? _ModalState$getState$ : null,

    /**
     * This function opens a modal based on the provided `modalName`.
     *
     * It will look at the stack passed to `<ModalProvider>` and add
     * the corresponding component to the current stack of open modals.
     * Alternatively, you can also provide some `params` that will be
     * accessible to that component.
     *
     * @example modalfy().openModal('PokédexEntryModal', { id: 619, name: 'Lin-Fu' }, () => console.log('PokédexEntryModal modal opened'))
     *
     * @see https://colorfy-software.gitbook.io/react-native-modalfy/api/types/modalprop#openmodal
     */
    openModal: (modalName, params, callback) => ModalState.openModal({
      modalName,
      params,
      callback,
      isCalledOutsideOfContext: true
    })
  };
};
export default ModalState;
//# sourceMappingURL=ModalState.js.map